'' Driver for Winbond SPI Flash Memory (W25X16AVDAIZ 2MB)
'' This is specifically for the TriBlade Prop Blade #1

'' P0  - DIO, SRAM   D0
'' P1  - DO,  SRAM   D1
'' P27 - /CS, SRAM   /CE, Mouse Dta
'' P28 - Clk, EEPROM SCL, Latch LE

CON DIOpin  =  0                       ' Data input / output pin
    DOpin   =  1                       ' Data output pin
    CSpin   = 27                       ' Chip select (active low)
    Clkpin  = 28                       ' Clock (active low)
  
    WrData  = $02_000000               ' Write Data / Address / Data
    RdData  = $03_000000               ' Read Data / Address / (Data)
    Status  = $05                      ' Read Status / (Data)
    WrtEna  = $06                      ' Write Enable
    SecEra  = $20_000000               ' Erase Sector / Address
    JEDEC   = $9F                      ' Return JEDEC device info
    JEDEC2M = $1530EF                  '  2MB Flash Memory (W25X16/16A)
    JEDEC4M = $1630EF                  '  4MB Flash Memory (W25X32)
    JEDEC8M = $1730EF                  '  8MB Flash Memory (W25X64)

    Busy    = $01                      ' Busy Status

    maxSeg  = $FE                      ' Last file segment (4K)
    secSize = 4096                     ' Number of bytes in sector
    secMask = secSize - 1              ' Mask for sector size

VAR long maxSize                       ' Size of Flash Memory

PUB start                              ' Initialize the object
   case sendRecv(JEDEC,3,0)            ' Get the JEDEC code for the device
      JEDEC2M: maxSize := 2 * |<20     ' Winbond W25X16/16A - 2MB Flash
      JEDEC4M: maxSize := 4 * |<20     ' Winbond W25X32     - 4MB Flash
      JEDEC8M: maxSize := 8 * |<20     ' Winbond W25X64     - 8MB Flash

PUB flashSize                          ' Return the size of the memory in bytes
   return maxSize

PUB sendRecv(sD, rCt, rA) | i, m       ' Send and possibly receive
'  sD  = Value to be sent MSB first, right justified in parameter
'  rCt = Number of bytes to be transferred (>0 received, <0 sent)
'  rA  = 0 or address of buffer area.  If 0, @RESULT is used
   outa[DIOpin]~
   dira[DIOpin]~~                      ' Set up the various I/O pins
   dira[DOpin]~
   outa[Clkpin]~
   dira[Clkpin]~~
   outa[CSpin]~
   dira[CSpin]~~
   if rA == 0                          ' Use RESULT if rA is zero
      rA := @result
   m := $FF000000
   repeat until m == 0 or sD & m <> 0  ' Set up sending mask
      m >>= 8
   m &= $80808080                      ' Start with most significant bit
   repeat until m == 0
      outa[DIOpin] := (sD & m) <> 0    ' Output bit to be sent
      outa[Clkpin]~~                   ' Toggle clock pulse
      outa[Clkpin]~
      m >>= 1                          ' Advance mask towards LSB
   if rCt < 0
      repeat i from 0 to -rCt-1
         m := $80                      ' Set up sending mask
         sD := byte[rA+i]              '  and get data byte
         repeat until m == 0
            outa[DIOpin] := (sD & m) <> 0
            outa[Clkpin]~~             ' Output bit to be set
            outa[Clkpin]~              ' Toggle clock pulse
            m >>= 1                    ' Advance mask towards LSB
   dira[DIOpin]~                       ' Don't need DIO pin now
   if rCt > 0
      repeat i from 0 to rCt-1
         m := $80                      ' Set up receiving mask
         sD~                           ' Make sure byte is zeroed
         repeat until m == 0
            if ina[DOpin] <> 0         ' Input bit to be received
               sD |= m
            outa[Clkpin]~~             ' Toggle clock pulse
            outa[Clkpin]~
            m >>= 1                    ' Advance mask towards LSB
         byte[rA+i] := sD
   outa[CSpin]~~                       ' All done, deselect device
   dira[DOpin]~
   dira[Clkpin]~                       ' Release the various I/O pins
   dira[CSpin]~

PUB readData(addr, data, count)
   sendRecv(RdData|addr,count,data)    ' Read data

PUB writeData(addr, data, count)
   sendRecv(WrtEna,0,0)                ' Enable writes
   sendRecv(WrData|addr,-count,data)   ' Write data
   repeat until sendRecv(Status,1,0) & Busy == 0 ' Wait until done

PUB eraseData(addr)
   sendRecv(WrtEna,0,0)                ' Enable writes
   sendRecv(SecEra|addr,0,0)           ' Erase 4K sector
   repeat until sendRecv(Status,1,0) & Busy == 0 ' Wait until done

PUB findFile(n1,n2,e,s) | i,a,b,c      ' Find match for file & seg #
   e := e << 8 | s
   repeat i from 0 to maxSize-secSize step secSize
      readData(i,@a,12)
      if a == n1 and b == n2 and c == e
         return i+12
   return -1

PUB findFree                           ' Find erased file segment
   return findFile($FFFFFFFF,$FFFFFFFF,$FFFFFF,$FF)
